// ------------------------------------------------------------------
// mesure-changement-governor - mesure de la duree du changement
//                              dynamique de governor.
//
// (c) 2011 Christophe BLAESS - Tous droits reserves.
//
// Exemple pour un post du blog http://www.blaess.fr/christophe/
// ------------------------------------------------------------------

#define _GNU_SOURCE // Pour avoir pthread_attr_setaffinity_np()

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>

// Parametrage du programme :
//
//   Le CPU (ou coeur) sur lequel le thread va s'executer
#    define CPU_THREAD               0
//   Le CPU (ou coeur) dont on modifie le governor
#    define CPU_MODIFIE              1
//   La priorite temps-reel du thread (0 = pas temps-reel)
#    define PRIORITE_THREAD          10

static void * thread (void * arg);
static char * governor_from = NULL;
static char * governor_to   = NULL;


int main(int argc, char * argv[])
{
	cpu_set_t  cpuset;
	pthread_t thr;
	pthread_attr_t attr;

	struct sched_param param;

	if (argc != 3) {
		fprintf(stderr, "usage: %s governor_1 governor_2\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	governor_from = argv[1];
	governor_to   = argv[2];
	
	pthread_attr_init(& attr);
	
	CPU_ZERO(& cpuset);
	CPU_SET(CPU_THREAD, & cpuset);
	if (pthread_attr_setaffinity_np(& attr, sizeof(cpuset), & cpuset) != 0) {
		perror("pthread_attr_setaffinity_np");
		exit(2);
	}
	if (PRIORITE_THREAD != 0) {
		if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO) != 0) {	
			perror("pthread_attr_setschedpolicy");
			exit(2);
		}
		param.sched_priority = PRIORITE_THREAD;
		if (pthread_attr_setschedparam(&attr, &param) != 0) {
			perror("pthread_attr_setschedparam");
			exit(2);
		}
		if (pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED) != 0) {
			perror("pthread_attr_setinheritsched");
			exit(2);
		}
	}
	if ((errno = pthread_create(& thr, &attr, thread, NULL)) != 0) {
		perror("pthread_create");
		exit(2);
	}
	pthread_join(thr, NULL);
	return EXIT_SUCCESS;
}

static void * thread (void * unused)
{
	int fd;
	char buffer[256];
	struct timeval avant, apres;
	long long int duree_1, duree_2;

	snprintf(buffer, 256, "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_governor", CPU_MODIFIE);

	fd = open(buffer, O_RDWR);
	if (fd < 0) {
		perror(buffer);
		exit(2);
	}
	
	snprintf(buffer, 256, "%s\n", governor_from);
	if (write(fd, buffer, strlen(buffer)) < 0) {
		perror(buffer);
		exit(2);
	}
	snprintf(buffer, 256, "%s\n", governor_to);
	// stabilisation du governor
	sleep(1);

	gettimeofday(& avant, NULL);
	if (write(fd, buffer, strlen(buffer)) < 0) {
		perror(buffer);
		exit(2);
	}
	gettimeofday(& apres, NULL);
	duree_1  = apres.tv_sec - avant.tv_sec;
	duree_1 *= 1000000;
	duree_1 += apres.tv_usec - avant.tv_usec;

	// duree due au gettimeofday()s
	gettimeofday(& avant, NULL);
	gettimeofday(& apres, NULL);
	duree_2  = apres.tv_sec - avant.tv_sec;
	duree_2 *= 1000000;
	duree_2 += apres.tv_usec - avant.tv_usec;

	close(fd);
	fprintf(stderr, "Duree de modification : %lldus\n",
		duree_1 - duree_2);
	pthread_exit(NULL);
}


